import itertools

import numpy as np

from .main import EvalCartesian


def write_single(filename, n, r, scaling, res=20):
    """This function creates a sphere mesh with "srgb1" values. Can be views in ParaView
    by disabling "Map Scalars".
    """
    import cplot
    import meshio
    import meshzoo

    points, cells = meshzoo.icosa_sphere(res)

    evaluator = EvalCartesian(points.T, scaling, complex_valued=True)
    vals = next(itertools.islice(evaluator, n, None))[r]

    srgb1_vals = cplot.get_srgb1(vals, colorspace="cam16")

    meshio.write_points_cells(
        filename, points, {"triangle": cells}, point_data={"srgb1": srgb1_vals}
    )


def write_tree(filename, n, scaling, res=20):
    import cplot
    import meshio
    import meshzoo

    points, cells = meshzoo.icosa_sphere(res)

    evaluator = EvalCartesian(points.T, scaling, complex_valued=True)
    meshes = []
    for L, level in enumerate(itertools.islice(evaluator, n)):
        for k, vals in enumerate(level):
            srgb1_vals = cplot.get_srgb1(vals, colorspace="cam16")
            # # exaggerate colors a bit
            # srgb1_vals *= 2.5
            # srgb1_vals[srgb1_vals > 1] = 1
            # srgb1_vals[srgb1_vals < 0] = 0

            pts = points.copy()

            pts[:, 0] += 2.2 * (k - L)
            pts[:, 2] -= 2.7 * L

            meshes.append(
                meshio.Mesh(pts, {"triangle": cells}, point_data={"srgb1": srgb1_vals})
            )

    # merge meshes
    points = np.concatenate([mesh.points for mesh in meshes])
    srgb1_vals = np.concatenate([mesh.point_data["srgb1"] for mesh in meshes])
    #
    cells = []
    k = 0
    for mesh in meshes:
        cells.append(mesh.cells[0].data + k)
        k += mesh.points.shape[0]
    cells = np.concatenate(cells)

    meshio.write_points_cells(
        filename, points, {"triangle": cells}, point_data={"srgb1": srgb1_vals}
    )



from . import gegenbauer


def plot(n, scaling):
    gegenbauer.plot(n, scaling, 0)


def show(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot(*args, **kwargs)
    plt.show()


def savefig(filename, *args, **kwargs):
    from matplotlib import pyplot as plt

    plot(*args, **kwargs)
    plt.savefig(filename, transparent=True, bbox_inches="tight")


class Eval:
    """Legendre. The first few are:

    scaling == "monic":
        1
        x
        x**2 - 1/3
        x**3 - 3*x/5
        x**4 - 6*x**2/7 + 3/35
        x**5 - 10*x**3/9 + 5*x/21

    scaling == "classical":
        1
        x
        3*x**2/2 - 1/2
        5*x**3/2 - 3*x/2
        35*x**4/8 - 15*x**2/4 + 3/8
        63*x**5/8 - 35*x**3/4 + 15*x/8

    scaling == "normal":
        sqrt(2)/2
        sqrt(6)*x/2
        3*sqrt(10)*x**2/4 - sqrt(10)/4
        5*sqrt(14)*x**3/4 - 3*sqrt(14)*x/4
        105*sqrt(2)*x**4/16 - 45*sqrt(2)*x**2/8 + 9*sqrt(2)/16
        63*sqrt(22)*x**5/16 - 35*sqrt(22)*x**3/8 + 15*sqrt(22)*x/16
    """

    def __init__(self, X, scaling, symbolic="auto"):
        self._gegenbauer_eval = gegenbauer.Eval(X, scaling, 0, symbolic)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._gegenbauer_eval)


class RecurrenceCoefficients(gegenbauer.RecurrenceCoefficients):
    def __init__(self, scaling, symbolic):
        super().__init__(scaling, 0, symbolic)



import numpy as np
import sympy


class Eval:
    """Evaluates orthogonal polynomials on the triangle.

    The return value is a list of arrays, where `out[k]` hosts the `2*k+1` values of the
    `k`th level of the tree

        (0, 0)
        (0, 1)   (1, 1)
        (0, 2)   (1, 2)   (2, 2)
          ...      ...      ...

    For reference, see

    Abedallah Rababah,
    Recurrence Relations for Orthogonal Polynomials on Triangular Domains,
    Mathematics 2016, 4(2), 25,
    <https://doi.org/10.3390/math4020025>.

    (The formulation there is more complicated than necessary, however, and doesn't
    include the normalization.)
    """

    def __init__(self, bary, scaling, symbolic="auto"):
        if symbolic == "auto":
            symbolic = np.asarray(bary).dtype == sympy.Basic

        self.bary = bary

        self.k = 0
        self.last = [None, None]

        self.rc = {"classical": RCClassical, "monic": RCMonic, "normal": RCNormal}[
            scaling
        ](symbolic)

        self.int_p0 = self.rc.p0

    def __iter__(self):
        return self

    def __next__(self):
        u, v, w = self.bary

        if self.k == 0:
            out = np.array([u * 0 + self.rc.p0])
        else:
            alpha, beta, gamma, delta, epsilon = self.rc[self.k]
            out = np.concatenate(
                [
                    self.last[0] * (np.multiply.outer(alpha, 1 - 2 * w).T - beta).T,
                    [delta * self.last[0][-1] * (u - v)],
                ]
            )

            if self.k > 1:
                out[:-2] -= (self.last[1].T * gamma).T
                out[-1] -= epsilon * self.last[1][-1] * (u + v) ** 2

        self.last[1] = self.last[0]
        self.last[0] = out
        self.k += 1
        return out


class RCClassical:
    def __init__(self, symbolic):
        self.S = sympy.S if symbolic else lambda x: x
        self.p0 = 1

    def __getitem__(self, n):
        n = self.S(n)

        r = np.arange(n)
        alpha = n * (2 * n + 1) / ((n - r) * (n + r + 1))
        beta = n * (2 * r + 1) ** 2 / ((n - r) * (n + r + 1) * (2 * n - 1))
        delta = (2 * n - 1) / n

        if n in [0, 1]:
            gamma = None
            epsilon = None
        else:
            r = np.arange(n - 1)
            gamma = (
                (n - r - 1)
                * (n + r)
                * (2 * n + 1)
                / ((n - r) * (n + r + 1) * (2 * n - 1))
            )
            epsilon = (n - 1) / n
        return alpha, beta, gamma, delta, epsilon


class RCMonic:
    def __init__(self, symbolic):
        self.S = sympy.S if symbolic else lambda x: x

        self.p0 = 1

    def __getitem__(self, n):
        assert n > 0

        r = np.arange(n)
        alpha = np.ones(n, dtype=int)
        delta = 1
        n = self.S(n)

        beta = (2 * r + 1) ** 2 / ((2 * n - 1) * (2 * n + 1))

        if n == 1:
            gamma = None
            epsilon = None
        else:
            r = np.arange(n - 1)
            gamma = (n - r - 1) ** 2 * (n + r) ** 2 / ((2 * n - 1) ** 2 * (n - 1) * n)
            epsilon = (n - 1) ** 2 / ((2 * n - 1) * (2 * n - 3))

        return alpha, beta, gamma, delta, epsilon


class RCNormal:
    """
    The coefficients here are based on the insight that

        int_T P_{n, r}^2 =
            int_0^1 L_r^2(t) dt * int_0^1 q_{n,r}(w)^2 (1-w)^(r+s+1) dw.

    For reference, see page 219 in

        Farouki, Goodman, Sauer,
        Construction of orthogonal bases for polynomials in Bernstein form on triangular
        and simplex domains,
        Computer Aided Geometric Design 20 (2003) 209–230,

    and the reference to Gould, 1972, there.

    The Legendre integral is 1/(2*r+1), and one gets

      int_T P_{n, r}^2 = 1 / (2*r+1) / (2*n+2)
                         * sum_{i=0}^{n-r} sum_{j=0}^{n-r}
                              (-1)**(i+j) * binom(n+r+1, i) * binom(n-r, i)
                                          * binom(n+r+1, j) * binom(n-r, j)
                                          / binom(2*n+1, i+j)

    Astonishingly, the double sum is always 1, hence

      int_T P_{n, r}^2 = 1 / (2*r+1) / (2*n+2).
    """

    def __init__(self, symbolic):
        self.S = sympy.S if symbolic else lambda x: x
        self.sqrt = sympy.sqrt if symbolic else np.sqrt

        self.p0 = self.sqrt(2)

    def __getitem__(self, n):
        n = self.S(n)
        sqrt = self.sqrt

        r = np.arange(n)
        alpha = sqrt((n + 1) * n) * ((2 * n + 1) / ((n - r) * (n + r + 1)))
        beta = (
            sqrt((n + 1) * n) * (2 * r + 1) ** 2 / ((n - r) * (n + r + 1) * (2 * n - 1))
        )
        delta = sqrt((2 * n + 1) * (n + 1) * (2 * n - 1) / n ** 3)

        if n in [0, 1]:
            gamma = None
            epsilon = None
        else:
            r = np.arange(n - 1)
            gamma = sqrt((n + 1) / (n - 1)) * (
                (n - r - 1)
                * (n + r)
                * (2 * n + 1)
                / ((n - r) * (n + r + 1) * (2 * n - 1))
            )
            epsilon = sqrt((2 * n + 1) * (n + 1) * (n - 1) / ((2 * n - 3) * n ** 2))

        return alpha, beta, gamma, delta, epsilon



from __future__ import annotations

import itertools

import numpy as np
import sympy

from ..helpers import Eval135


def plot(n: int, *args, **kwargs):
    import matplotx
    from matplotlib import pyplot as plt

    plt.style.use(matplotx.styles.dufte)

    x = np.linspace(-1.0, 1.0, 200)
    for k, level in enumerate(itertools.islice(Eval(x, *args, **kwargs), n + 1)):
        # Choose all colors in each level approximately equal, around the reference
        # color.
        ref_color = plt.rcParams["axes.prop_cycle"].by_key()["color"][k]
        ref_rgb = np.array(
            list(int(ref_color[1:][i : i + 2], 16) / 255 for i in (0, 2, 4))
        )
        for l, entry in enumerate(level):
            col = ref_rgb * (1 + (l - k) / (2 * (k + 1)))
            col[col < 0.0] = 0.0
            col[col > 1.0] = 1.0
            plt.plot(x, entry, label=f"n={k}, r={l - k}", color=col)

    plt.grid(axis="x")
    matplotx.line_labels()
    ax = plt.gca()
    ax.spines["top"].set_visible(False)
    ax.spines["bottom"].set_visible(False)

    (scaling,) = args
    plt.title(f'Associated Legendre "polynomials" (scaling={scaling})')
    ax.spines["right"].set_visible(True)
    ax.spines["left"].set_visible(True)
    plt.xlim(-1.0, 1.0)


def show(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot(*args, **kwargs)
    plt.show()


def savefig(filename, *args, **kwargs):
    from matplotlib import pyplot as plt

    plot(*args, **kwargs)
    plt.savefig(filename, transparent=True, bbox_inches="tight")


class Eval:
    """
    Useful references are

    Taweetham Limpanuparb, Josh Milthorpe,
    Associated Legendre Polynomials and Spherical Harmonics Computation for Chemistry
    Applications,
    Proceedings of The 40th Congress on Science and Technology of Thailand;
    2014 Dec 2-4, Khon Kaen, Thailand. P. 233-241.
    <https://arxiv.org/abs/1410.1748>

    and

    Schneider et al.,
    A new Fortran 90 program to compute regular and irregular associated Legendre
    functions,
    Computer Physics Communications,
    Volume 181, Issue 12, December 2010, Pages 2091-2097,
    <https://doi.org/10.1016/j.cpc.2010.08.038>.
    """

    def __init__(self, X, scaling: str, symbolic: str | bool = "auto"):
        cls = {"classical": RCClassical, "normal": RCNormal}[scaling]
        if symbolic == "auto":
            symbolic = np.asarray(X).dtype == sympy.Basic
        rc = cls(symbolic)
        self._eval135 = Eval135(rc, X, symbolic=symbolic)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._eval135)


class RCClassical:
    def __init__(self, symbolic: bool):
        self.frac = sympy.Rational if symbolic else lambda x, y: x / y
        self.p0 = 1

    def __getitem__(self, L: int):
        z0 = self.frac(1, 2 * L)
        z1 = -(2 * L - 1)
        c0 = [self.frac(2 * L - 1, L - m) for m in range(-L + 1, L)]
        if L == 1:
            c1 = None
        else:
            c1 = [self.frac(L - 1 + m, L - m) for m in range(-L + 2, L - 1)]
        return z0, z1, c0, c1


class RCNormal:
    def __init__(self, symbolic: bool):
        self.sqrt = np.vectorize(sympy.sqrt) if symbolic else np.sqrt
        self.frac = np.vectorize(sympy.Rational) if symbolic else lambda x, y: x / y

        self.p0 = 1 / self.sqrt(2)

    def __getitem__(self, L: int):
        z0 = self.sqrt(self.frac(2 * L + 1, 2 * L))
        z1 = -self.sqrt(self.frac(2 * L + 1, 2 * L))
        #
        m = np.arange(-L + 1, L)
        c0 = self.sqrt(self.frac((2 * L - 1) * (2 * L + 1), (L + m) * (L - m)))
        #
        if L == 1:
            c1 = None
        else:
            m = np.arange(-L + 2, L - 1)
            c1 = self.sqrt(
                self.frac(
                    (L + m - 1) * (L - m - 1) * (2 * L + 1),
                    (2 * L - 3) * (L + m) * (L - m),
                )
            )
        return z0, z1, c0, c1



import itertools

import numpy as np

from .main import Eval


def plot(n, *args, **kwargs):
    import matplotx
    from matplotlib import pyplot as plt

    plt.style.use(matplotx.styles.dufte)

    x = np.linspace(0.0, 4.4, 100)
    for k, level in enumerate(itertools.islice(Eval(x, *args, **kwargs), n + 1)):
        plt.plot(x, level, label=f"n={k}")

    plt.grid(axis="x")
    matplotx.line_labels()
    ax = plt.gca()

    (scaling,) = args
    alpha = kwargs["alpha"]
    plt.title(f"(Generalized) Laguerre polynomials (α={alpha}, scaling={scaling})")
    ax.spines["left"].set_visible(True)


def show(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot(*args, **kwargs)
    plt.show()


def savefig(filename, *args, **kwargs):
    from matplotlib import pyplot as plt

    plot(*args, **kwargs)
    plt.savefig(filename, transparent=True, bbox_inches="tight")



from __future__ import annotations

import numpy as np
import sympy
from numpy.typing import ArrayLike

try:
    # Python 3.8+
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def savefig_single(filename, *args, **kwargs):
    from matplotlib import pyplot as plt

    plot_single(*args, **kwargs)
    plt.savefig(filename, transparent=True, bbox_inches="tight")


def show_single(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot_single(*args, **kwargs)
    plt.show()


def plot_single(*args, **kwargs):
    from .tools import plot_single as ps

    ps("Xu", Eval, *args, **kwargs)


def savefig_tree(filename, *args, **kwargs):
    from matplotlib import pyplot as plt

    plot_tree(*args, **kwargs)
    plt.savefig(filename, transparent=True, bbox_inches="tight")


def show_tree(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot_tree(*args, **kwargs)
    plt.show()


def plot_tree(*args, **kwargs):
    from .tools import plot_tree as pt

    pt("Xu", Eval, *args, **kwargs)


class Eval:
    """Evaluates the entire tree of orthogonal polynomials on the unit disk.

    The return value is a list of arrays, where `out[k]` hosts the `2*k+1`
    values of the `k`th level of the tree

        (0, 0)
        (0, 1)   (1, 1)
        (0, 2)   (1, 2)   (2, 2)
          ...      ...      ...

    See

    Yuan Xu
    Orthogonal polynomials of several variables,
    Jan. 2017,
    <https://arxiv.org/abs/1701.02709>

    equation (3.4) for a formulation in terms of Gegenbauer polynomials C. The
    recurrence relation can be worked out from there.
    """

    def __init__(
        self,
        X: ArrayLike,
        scaling: Literal["classical"] | Literal["monic"] | Literal["normal"],
        symbolic: Literal["auto"] | bool = "auto",
    ):
        if symbolic == "auto":
            symbolic = np.asarray(X).dtype == sympy.Basic

        self.rc = {"classical": RCClassical, "monic": RCMonic, "normal": RCNormal}[
            scaling
        ](symbolic)

        self.X = X
        self.one_min_x2 = 1 - self.X[0] ** 2
        self.L = 0
        self.last = [None, None]

        pi = sympy.pi if symbolic else np.pi
        self.int_p0 = self.rc.p0 * pi

    def __iter__(self):
        return self

    def __next__(self):
        if self.L == 0:
            out = np.array([0 * self.X[0] + self.rc.p0])
        else:
            alpha, beta, gamma, delta = self.rc[self.L]
            out = np.concatenate(
                [
                    self.last[0] * np.multiply.outer(alpha, self.X[0]),
                    [self.last[0][-1] * beta * self.X[1]],
                ]
            )

            if self.L > 1:
                out[:-2] -= (self.last[1].T * gamma).T
                out[-1] -= self.last[1][-1] * delta * self.one_min_x2

        self.last[1] = self.last[0]
        self.last[0] = out
        self.L += 1
        return out


class RCClassical:
    """The maximum values (which are attained at (1, 0) for the first and (0, 1) for the
    last polynomial in each level) is 1.
    """

    def __init__(self, symbolic: bool, mu: int | float = 1):
        self.S = sympy.S if symbolic else lambda x: x
        self.sqrt = np.vectorize(sympy.sqrt) if symbolic else np.sqrt
        self.mu = mu
        assert self.mu > -1

        self.p0 = 1

    # How to get from the monic representation to this one:
    # The value at (0, 1) needs to be 1, so
    #
    # self.last[0][-1] * beta * self.X[1] - self.last[1][-1] * delta * self.one_min_x2
    # = beta - delta
    #
    # (from the recurrence) needs to be 1. Hence
    #
    #  beta_new = beta / (beta - delta)
    #  delta_new = delta / (beta - delta) / beta_min_delta_from_previous_step
    #
    # Calling z = beta - delta, this establishes the formula
    #
    #     z(n) = beta(n) - delta(n) / z(n-1)
    #
    # which is a continued fraction. Using sympy, it can be shown that
    #
    #   z(n) = (mu + n - 1) / (mu + 2 * (n - 1))
    #
    # so
    #
    #   beta(n) = (mu + 2 * (n - 1)) / (mu + n - 1)
    #
    # and delta accordingly. The same reasoning can be applied to alpha and gamma, the
    # result will be
    #
    #   z(n) = (mu + n) / (mu + 2 * n - 1)
    #
    # TODO only the leftmost and rightmost values are correct
    def __getitem__(self, n: int):
        assert n > 0

        n = self.S(n)
        mu = self.mu

        alpha = np.full(n, (mu + 2 * n - 1) / (mu + n))
        # case distinction for mu==0
        if n == 1:
            beta = 1
        else:
            beta = (mu + 2 * n - 2) / (mu + n - 1)

        if n == 1:
            gamma = None
            delta = None
        else:  # n > 1
            k = np.arange(n - 1)
            gamma = (n - 1 - k) * (n + k + mu - 1) / ((mu + n) * (mu + n - 1))
            delta = (n - 1) / (mu + n - 1)

        return alpha, beta, gamma, delta


class RCMonic:
    """alpha and beta both equal 1."""

    def __init__(self, symbolic: bool, mu: int | float = 1):
        self.S = sympy.S if symbolic else lambda x: x
        self.sqrt = np.vectorize(sympy.sqrt) if symbolic else np.sqrt
        self.mu = mu
        assert self.mu > -1

        self.p0 = 1

    def __getitem__(self, n: int):
        assert n > 0

        n = self.S(n)
        mu = self.mu

        alpha = np.ones(n, dtype=int)
        beta = 1

        if n == 1:
            gamma = None
            delta = None
        else:
            k = np.arange(n - 1)
            gamma = (
                (n - 1 - k) * (n + k + mu - 1) / ((2 * n + mu - 3) * (2 * n + mu - 1))
            )
            # case distinction to avoid undefined expression for mu = 0.
            if n == 2:
                delta = self.S(1) / (mu + 2)
            else:
                delta = (n - 1) * (n + mu - 2) / ((2 * n + mu - 2) * (2 * n + mu - 4))

        return alpha, beta, gamma, delta


class RCNormal:
    """Recurrence coefficients for the disk with the Gegenbauer-style weight function
    $$
       W_{\\mu}(x, y) = \\frac{\\mu + 1}{2\\pi} (1 - x^2 - y^2)^{(\\mu - 1)/2}
    $$
    scaled for normality. The default choice is \\mu=1, giving W_1(x, y) = 1.

    In some other tests, one finds a rescaled mu.
    $$
    \\frac{\\mu + 1/2}{\\pi} (1 - x^2 - y^2)^{\\mu - 1/2}
    $$

    This is based on the first orthonormal basis in section 3.2 of

    Yuan Xu,
    Orthogonal polynomials of several variables,
    <https://arxiv.org/pdf/1701.02709.pdf>,

    specifically equation (3.4)

      P_k^n(x, y) = h_{k,}^{-1}
                    C_{n-k}^{k+\\mu+1/2}(x)
                    \\sqrt{1-x^2}^k C_k^{\\mu}(y/\\sqrt(1-x^2)

    with C_n^{\\lambda} being the Gegenbauer polynomial, scaled such that
    $C_n^{\\lambda}(1)=\\frac{(\\lambda+1)_n}{n!}$ and

      h_{k,n}^2 = \\frac{(2k+2\\mu+1)_{n-k} (2\\mu)_k (\\mu)_k (\\mu+1/2)}
                        {(n-k)!k! (\\mu+1/2)_k (n+\\mu+1/2)}.

    The recurrence coefficients are retrieved by exploiting the Gegenbauer recurrence

       C_n^{\\lambda}(t) =
           1/n (
               + 2t (n+\\lambda+1) C_{n-1}^{\\lambda}(t)
               - (n+2\\lambda-2) C_{n-2}^{\\lambda}(t)
               )

    One gets

        P_k^n(x, y) = + 2 \\alpha_{k,n} x P_k^{n-1}(x, y)
                      - \\beta_{k, n} P_k^{n-2}(x, y)

    with

        \\alpha_{k, n}^2 = \\frac{(n+\\mu+1/2)(n+\\mu-1/2)}{(n-k)(n+k+2\\mu)},
        \\beta_{k, n}^2 = \\frac{(n-k-1) (n+\\mu+1/2)(n+k+2\\mu-1)}
                                {(n-k)(n+\\mu-3/2)(n+k+2\\mu)},

    and

        P_n^n(x, y) = + 2 \\gamma_{k,n} y P_{n-1}^{n-1}(x, y)
                      - \\delta_{k, n} (1-x^2) P_{n-1}^{n-2}(x, y)

    with

        \\gamma_{k, n}^2 = \\frac{(n+\\mu-1)(n+\\mu+1/2)}{n (n+2\\mu-1)},
        \\delta_{k, n}^2 = \\frac{(n+2\\mu-2) (n-1) (n+\\mu-1/2) (n+\\mu+1/2)}
                                 {n (n+2\\mu-1) (n+\\mu-1) (n+\\mu-2)}.
    """

    # default: weight function 1
    def __init__(self, symbolic, mu=1):
        self.S = sympy.S if symbolic else lambda x: x
        self.sqrt = np.vectorize(sympy.sqrt) if symbolic else np.sqrt
        self.mu = mu
        assert self.mu > -1

        pi = sympy.pi if symbolic else np.pi
        self.p0 = 1 / self.sqrt(pi)

    def __getitem__(self, n: int):
        assert n > 0

        n = self.S(n)
        mu = self.mu

        k = np.arange(n)
        alpha = self.sqrt(
            (2 * n + mu + 1) * (2 * n + mu - 1) / ((n - k) * (n + k + mu))
        )

        beta = self.sqrt((2 * n + mu - 2) * (2 * n + mu + 1) / (n * (n + mu - 1)))

        if n == 1:
            gamma = None
            delta = None
        else:
            k = np.arange(n - 1)
            gamma = self.sqrt(
                (n - 1 - k)
                * (2 * n + mu + 1)
                * (n + k + mu - 1)
                / ((n - k) * (2 * n + mu - 3) * (n + k + mu))
            )
            # case distinction to avoid undefined expression for mu = 0.
            if n == 2:
                two = self.S(2)
                delta = self.sqrt((3 + mu) * (5 + mu) / (two * (1 + mu) * (2 + mu)))
            else:
                delta = self.sqrt(
                    (n - 1)
                    * (n + mu - 2)
                    * (2 * n + mu - 1)
                    * (2 * n + mu + 1)
                    / (n * (n + mu - 1) * (2 * n + mu - 2) * (2 * n + mu - 4))
                )
        return alpha, beta, gamma, delta



import math

import numpy as np
import sympy

from ..helpers import Eval1D


def plot(n, *args, **kwargs):
    import matplotx
    from matplotlib import pyplot as plt

    plt.style.use(matplotx.styles.dufte)

    x = np.linspace(-1.0, 1.0, 100)
    evaluator = Eval(x, *args, **kwargs)
    for k in range(n + 1):
        plt.plot(x, next(evaluator), label=f"n={k}")

    plt.grid(axis="x")
    matplotx.line_labels()
    ax = plt.gca()

    alpha, beta, scaling = args
    if alpha == beta:
        if alpha == 0:
            plt.title(f"Legendre polynomials (scaling={scaling})")
        elif alpha == -0.5:
            plt.title(f"Chebyshev 1 polynomials (scaling={scaling})")
        elif alpha == +0.5:
            plt.title(f"Chebyshev 2 polynomials (scaling={scaling})")
        else:
            plt.title(f"Gegenbauer polynomials (λ={alpha}, scaling={scaling})")
    else:
        plt.title(f"Jacobi polynomials (α={alpha}, β={beta}, scaling={scaling})")
    ax.spines["right"].set_visible(True)
    ax.spines["left"].set_visible(True)


def show(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot(*args, **kwargs)
    plt.show()


def savefig(filename, *args, **kwargs):
    from matplotlib import pyplot as plt

    plot(*args, **kwargs)
    plt.savefig(filename, transparent=True, bbox_inches="tight")


class Eval:
    def __init__(self, X, *args, symbolic="auto"):
        if symbolic == "auto":
            symbolic = np.asarray(X).dtype == sympy.Basic

        self._eval_1d = Eval1D(X, RecurrenceCoefficients(*args, symbolic=symbolic))

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._eval_1d)


class RecurrenceCoefficients:
    def __init__(self, scaling, alpha, beta, symbolic):
        cls = {"monic": _RCMonic, "classical": _RCClassical, "normal": _RCNormal}[
            scaling
        ]
        self.rc = cls(alpha, beta, symbolic)
        self.p0 = self.rc.p0

        gamma = sympy.gamma if symbolic else lambda x: math.gamma(float(x))
        self.int_1 = (
            2 ** (alpha + beta + 1)
            * gamma(alpha + 1)
            * gamma(beta + 1)
            / gamma(alpha + beta + 2)
        )

    def __getitem__(self, N):
        return self.rc[N]


class _RCMonic:
    """Generate the recurrence coefficients a_k, b_k, c_k in

    P_{k+1}(x) = (a_k x - b_k) * P_{k}(x) - c_k * P_{k-1}(x)

    for the Jacobi polynomials which are orthogonal on [-1, 1] with respect to the
    weight w(x)=[(1-x)^alpha]*[(1+x)^beta]; see
    <https://en.wikipedia.org/wiki/Jacobi_polynomials#Recurrence_relations>.
    """

    def __init__(self, alpha, beta, symbolic):
        self.alpha = alpha
        self.beta = beta
        self.gamma = sympy.gamma if symbolic else lambda x: math.gamma(float(x))

        self.frac = sympy.Rational if symbolic else lambda x, y: x / y
        self.nan = None if symbolic else math.nan
        self.one = 1 if symbolic else 1.0

        self.p0 = self.one

    def __getitem__(self, N):
        frac = self.frac
        alpha = self.alpha
        beta = self.beta

        a = self.one

        if N == 0:
            b = frac(beta - alpha, alpha + beta + 2)
        else:
            b = frac(
                beta ** 2 - alpha ** 2,
                (2 * N + alpha + beta) * (2 * N + alpha + beta + 2),
            )

        # Note that we have the treat the case N==1 separately to avoid division by 0
        # for alpha=beta=-1/2.
        if N == 0:
            # c[0] is not used in the actual recurrence, but is traditionally defined as
            # the integral of the weight function of the domain, i.e.,
            # ```
            # int_{-1}^{+1} (1-x)^a * (1+x)^b dx =
            #     2^(a+b+1) * Gamma(a+1) * Gamma(b+1) / Gamma(a+b+2).
            # ```
            # This is bad practice; the value could accidentally be used.
            c = self.nan
        elif N == 1:
            c = frac(
                4 * (1 + alpha) * (1 + beta),
                (2 + alpha + beta) ** 2 * (3 + alpha + beta),
            )
        else:
            c = frac(
                4 * (N + alpha) * (N + beta) * N * (N + alpha + beta),
                (2 * N + alpha + beta) ** 2
                * (2 * N + alpha + beta + 1)
                * (2 * N + alpha + beta - 1),
            )
        return a, b, c


class _RCClassical:
    def __init__(self, alpha, beta, symbolic):
        self.frac = sympy.Rational if symbolic else lambda x, y: x / y
        self.nan = None if symbolic else math.nan
        self.alpha = alpha
        self.beta = beta

        self.p0 = 1 if symbolic else 1.0

    def __getitem__(self, N):
        frac = self.frac
        alpha = self.alpha
        beta = self.beta

        # Treat N = 0 separately to avoid division by 0 for alpha = beta = -1/2.
        if N == 0:
            a = frac(alpha + beta + 2, 2)
            b = frac(beta - alpha, 2)
            c = self.nan
        else:
            a = frac(
                (2 * N + alpha + beta + 1) * (2 * N + alpha + beta + 2),
                2 * (N + 1) * (N + alpha + beta + 1),
            )
            b = frac(
                (beta ** 2 - alpha ** 2) * (2 * N + alpha + beta + 1),
                2 * (N + 1) * (N + alpha + beta + 1) * (2 * N + alpha + beta),
            )
            c = frac(
                (N + alpha) * (N + beta) * (2 * N + alpha + beta + 2),
                (N + 1) * (N + alpha + beta + 1) * (2 * N + alpha + beta),
            )

        return a, b, c


class _RCNormal:
    def __init__(self, alpha, beta, symbolic):
        self.frac = sympy.Rational if symbolic else lambda x, y: x / y
        self.sqrt = sympy.sqrt if symbolic else math.sqrt
        self.nan = None if symbolic else math.nan
        self.alpha = alpha
        self.beta = beta

        gamma = sympy.gamma if symbolic else lambda x: math.gamma(float(x))
        self.int_1 = (
            2 ** (alpha + beta + 1)
            * gamma(alpha + 1)
            * gamma(beta + 1)
            / gamma(alpha + beta + 2)
        )
        self.p0 = self.sqrt(1 / self.int_1)

    def __getitem__(self, N):
        frac = self.frac
        sqrt = self.sqrt
        alpha = self.alpha
        beta = self.beta

        # Treat N==0 separately to avoid division by 0 for alpha=beta=-1/2 (Chebyshev 1)
        # and alpha=beta=0 (Legendre).
        if N == 0:
            t = sqrt(frac(alpha + beta + 3, (alpha + 1) * (beta + 1)))
            a = frac(alpha + beta + 2, 2) * t
            b = frac(beta - alpha, 2) * t
        else:
            t = sqrt(
                frac(
                    (2 * N + alpha + beta + 1) * (2 * N + alpha + beta + 3),
                    (N + 1) * (N + alpha + 1) * (N + beta + 1) * (N + alpha + beta + 1),
                )
            )
            a = frac(2 * N + alpha + beta + 2, 2) * t
            b = frac(beta ** 2 - alpha ** 2, 2 * (2 * N + alpha + beta)) * t

        if N == 0:
            c = self.nan
        elif N == 1:
            c = frac(4 + alpha + beta, 2 + alpha + beta) * sqrt(
                frac(
                    (1 + alpha) * (1 + beta) * (5 + alpha + beta),
                    2 * (2 + alpha) * (2 + beta) * (2 + alpha + beta),
                )
            )
        else:
            c = frac(2 * N + alpha + beta + 2, 2 * N + alpha + beta) * sqrt(
                frac(
                    N
                    * (N + alpha)
                    * (N + beta)
                    * (N + alpha + beta)
                    * (2 * N + alpha + beta + 3),
                    (N + 1)
                    * (N + alpha + 1)
                    * (N + beta + 1)
                    * (N + alpha + beta + 1)
                    * (2 * N + alpha + beta - 1),
                )
            )

        return a, b, c



import itertools

import numpy as np

from .main import Eval


def savefig_single(filename, *args, **kwargs):
    from matplotlib import pyplot as plt

    plot_single(*args, **kwargs)
    plt.savefig(filename, transparent=True, bbox_inches="tight")


def show_single(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot_single(*args, **kwargs)
    plt.show()


def plot_single(
    degrees, res=100, scaling="normal", colorbar=True, cmap="RdBu_r", corners=None
):
    import meshzoo
    from matplotlib import pyplot as plt

    n = sum(degrees)
    r = degrees[0]

    def f(bary):
        for k, level in enumerate(Eval(bary, scaling)):
            if k == n:
                return level[r]

    if corners is None:
        alpha = np.pi * np.array([7.0 / 6.0, 11.0 / 6.0, 3.0 / 6.0])
        corners = np.array([np.cos(alpha), np.sin(alpha)])

    bary, cells = meshzoo.triangle(res)
    x, y = np.dot(corners, bary)
    z = np.array(f(bary), dtype=float)

    plt.tripcolor(x, y, cells, z, shading="flat")

    if colorbar:
        plt.colorbar()
    # Choose a diverging colormap such that the zeros are clearly distinguishable.
    plt.set_cmap(cmap)
    # Make sure the color map limits are symmetric around 0.
    clim = plt.gci().get_clim()
    mx = max(abs(clim[0]), abs(clim[1]))
    plt.clim(-mx, mx)

    # triangle outlines
    X = np.column_stack([corners, corners[:, 0]])
    plt.plot(X[0], X[1], "-k")

    plt.gca().set_aspect("equal")
    plt.axis("off")
    plt.title(
        f"Orthogonal polynomial on triangle ([{degrees[0]}, {degrees[1]}], {scaling})"
    )


def savefig_tree(filename, *args, **kwargs):
    from matplotlib import pyplot as plt

    plot_tree(*args, **kwargs)
    plt.savefig(filename, transparent=True, bbox_inches="tight")


def show_tree(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot_tree(*args, **kwargs)
    plt.show()


# Use a diverging colormap by default so the zeros are well recognizable
# https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
def plot_tree(n, res=100, scaling="normal", colorbar=False, cmap="RdBu_r", clim=None):
    import matplotx
    import meshzoo
    from matplotlib import pyplot as plt

    plt.style.use(matplotx.styles.dufte)

    bary, cells = meshzoo.triangle(res)
    evaluator = Eval(bary, scaling)

    plt.set_cmap(cmap)
    plt.gca().set_aspect("equal")
    plt.axis("off")

    for k, level in enumerate(itertools.islice(evaluator, n + 1)):
        for r, z in enumerate(level):
            alpha = np.pi * np.array([7.0 / 6.0, 11.0 / 6.0, 3.0 / 6.0])
            corners = np.array([np.cos(alpha), np.sin(alpha)])
            corners[0] += 2.1 * (r - k / 2)
            corners[1] -= 1.9 * k
            x, y = np.dot(corners, bary)

            plt.tripcolor(x, y, cells, z, shading="flat")
            plt.clim(clim)

            # triangle outlines
            X = np.column_stack([corners, corners[:, 0]])
            plt.plot(X[0], X[1], "-k")

    if colorbar:
        plt.colorbar()



import itertools

import numpy as np


def plot_single(
    name,
    evaluator,
    degrees,
    res=50,
    scaling="normal",
    colorbar=True,
    cmap="RdBu_r",
):
    import meshzoo
    from matplotlib import pyplot as plt

    n = sum(degrees)
    r = degrees[0]

    def f(bary):
        for k, level in enumerate(evaluator(bary, scaling)):
            if k == n:
                return level[r]

    points, cells = meshzoo.disk(6, res)
    z = np.array(f(points.T), dtype=float)

    plt.tripcolor(points[:, 0], points[:, 1], cells, z, shading="flat")

    if colorbar:
        plt.colorbar()
    # Choose a diverging colormap such that the zeros are clearly distinguishable.
    plt.set_cmap(cmap)
    # Make sure the color map limits are symmetric around 0.
    clim = plt.gci().get_clim()
    mx = max(abs(clim[0]), abs(clim[1]))
    plt.clim(-mx, mx)

    # circle outline
    circle = plt.Circle((0, 0), 1.0, edgecolor="k", fill=False)
    plt.gca().add_patch(circle)

    plt.gca().set_aspect("equal")
    plt.axis("off")
    plt.title(
        f"{name} orthogonal polynomial on disk "
        f"([{degrees[0]}, {degrees[1]}], {scaling})"
    )


def savefig_tree(filename, *args, dpi=None, **kwargs):
    from matplotlib import pyplot as plt

    plot_tree(*args, **kwargs)
    plt.savefig(filename, dpi=dpi, transparent=True, bbox_inches="tight")


def show_tree(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot_tree(*args, **kwargs)
    plt.show()


# Use a diverging colormap by default so the zeros are well recognizable
# https://matplotlib.org/3.1.0/tutorials/colors/colormaps.html
def plot_tree(
    name,
    evaluator,
    n,
    res=50,
    scaling="normal",
    colorbar=False,
    cmap="RdBu_r",
    clim=None,
    show_title=True,
):
    import matplotx
    import meshzoo
    from matplotlib import pyplot as plt

    plt.style.use(matplotx.styles.dufte)

    points, cells = meshzoo.disk(6, res)
    evaluator = evaluator(points.T, scaling)

    plt.set_cmap(cmap)
    plt.gca().set_aspect("equal")
    plt.axis("off")

    for k, level in enumerate(itertools.islice(evaluator, n + 1)):
        for r, z in enumerate(level):
            pts = points.copy()
            offset = [2.6 * (r - k / 2), -2.3 * k]

            pts[:, 0] += offset[0]
            pts[:, 1] += offset[1]

            plt.tripcolor(pts[:, 0], pts[:, 1], cells, z, shading="flat")
            plt.clim(clim)

            # circle outline
            circle = plt.Circle(offset, 1.0, edgecolor="k", fill=False)
            plt.gca().add_patch(circle)

    if colorbar:
        plt.colorbar()

    if show_title:
        plt.title(f"{name} orthogonal polynomials on the disk ({scaling})")



from .main import Eval
from .tools import (
    plot_single,
    plot_tree,
    savefig_single,
    savefig_tree,
    show_single,
    show_tree,
)

__all__ = [
    "Eval",
    "plot_single",
    "show_single",
    "savefig_single",
    "plot_tree",
    "savefig_tree",
    "show_tree",
]



from __future__ import annotations

try:
    # Python 3.8+
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import numpy as np
import sympy

from . import gegenbauer


def plot(n, scaling):
    gegenbauer.plot(n, scaling, +0.5)


def show(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot(*args, **kwargs)
    plt.show()


def savefig(filename, *args, **kwargs):
    from matplotlib import pyplot as plt

    plot(*args, **kwargs)
    plt.savefig(filename, transparent=True, bbox_inches="tight")


class Eval:
    """Chebyshev polynomials of the second kind. The first few are:

    scaling == "monic":
        1
        x
        x**2 - 1/4
        x**3 - x/2
        x**4 - 3*x**2/4 + 1/16
        x**5 - x**3 + 3*x/16

    scaling == "classical":
        1
        3*x/2
        5*x**2/2 - 5/8
        35*x**3/8 - 35*x/16
        63*x**4/8 - 189*x**2/32 + 63/128
        231*x**5/16 - 231*x**3/16 + 693*x/256

    scaling == "normal":
        sqrt(2)/sqrt(pi)
        2*sqrt(2)*x/sqrt(pi)
        4*sqrt(2)*x**2/sqrt(pi) - sqrt(2)/sqrt(pi)
        8*sqrt(2)*x**3/sqrt(pi) - 4*sqrt(2)*x/sqrt(pi)
        16*sqrt(2)*x**4/sqrt(pi) - 12*sqrt(2)*x**2/sqrt(pi) + sqrt(2)/sqrt(pi)
        32*sqrt(2)*x**5/sqrt(pi) - 32*sqrt(2)*x**3/sqrt(pi) + 6*sqrt(2)*x/sqrt(pi)
    """

    def __init__(self, X, scaling: str, symbolic: Literal["auto"] | bool = "auto"):
        if symbolic == "auto":
            symbolic = np.asarray(X).dtype == sympy.Basic

        lmbda = sympy.S(1) / 2 if symbolic else 0.5
        self._gegenbauer_eval = gegenbauer.Eval(X, scaling, lmbda, symbolic)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._gegenbauer_eval)



import numpy as np
import sympy

from ..c1 import jacobi
from ..helpers import ProductEval, ProductEvalWithDegrees


class Eval:
    def __init__(self, X, alpha=0, beta=0, symbolic="auto", return_degrees=False):
        if symbolic == "auto":
            symbolic = np.asarray(X).dtype == sympy.Basic

        rc = jacobi.RecurrenceCoefficients("normal", alpha, beta, symbolic)
        cls = ProductEvalWithDegrees if return_degrees else ProductEval
        self._product_eval = cls(rc, 1, X)
        self.int_p0 = self._product_eval.int_p0

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._product_eval)



import numpy as np


def clenshaw(a, alpha, beta, t):
    """Clenshaw's algorithm for evaluating

    S(t) = \\sum a_k P_k(alpha, beta)(t)

    where P_k(alpha, beta) is the kth orthogonal polynomial defined by the
    recurrence coefficients alpha, beta.

    See <https://en.wikipedia.org/wiki/Clenshaw_algorithm> for details.
    """
    n = len(alpha)
    assert len(beta) == n
    assert len(a) == n + 1

    try:
        b = np.empty((n + 1,) + t.shape)
    except AttributeError:  # 'float' object has no attribute 'shape'
        b = np.empty(n + 1)

    # b[0] is unused, can be any value
    # TODO shift the array
    b[0] = 1.0

    b[n] = a[n]
    b[n - 1] = a[n - 1] + (t - alpha[n - 1]) * b[n]
    for k in range(n - 2, 0, -1):
        b[k] = a[k] + (t - alpha[k]) * b[k + 1] - beta[k + 1] * b[k + 2]

    phi0 = 1.0
    phi1 = t - alpha[0]

    return phi0 * a[0] + phi1 * b[1] - beta[1] * phi0 * b[2]



import numpy as np
import sympy

from . import gegenbauer


def plot(n, scaling):
    gegenbauer.plot(n, scaling, -0.5)


def show(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot(*args, **kwargs)
    plt.show()


def savefig(filename, *args, **kwargs):
    from matplotlib import pyplot as plt

    plot(*args, **kwargs)
    plt.savefig(filename, transparent=True, bbox_inches="tight")


class Eval:
    """Chebyshev polynomials of the first kind. The first few are:

    scaling == "monic":
        1
        x
        x**2 - 1/2
        x**3 - 3*x/4
        x**4 - x**2 + 1/8
        x**5 - 5*x**3/4 + 5*x/16

    scaling == "classical":
        1
        x/2
        3*x**2/4 - 3/8
        5*x**3/4 - 15*x/16
        35*x**4/16 - 35*x**2/16 + 35/128
        63*x**5/16 - 315*x**3/64 + 315*x/256

    scaling == "normal":
        1/sqrt(pi)
        sqrt(2)*x/sqrt(pi)
        2*sqrt(2)*x**2/sqrt(pi) - sqrt(2)/sqrt(pi)
        4*sqrt(2)*x**3/sqrt(pi) - 3*sqrt(2)*x/sqrt(pi)
        8*sqrt(2)*x**4/sqrt(pi) - 8*sqrt(2)*x**2/sqrt(pi) + sqrt(2)/sqrt(pi)
        16*sqrt(2)*x**5/sqrt(pi) - 20*sqrt(2)*x**3/sqrt(pi) + 5*sqrt(2)*x/sqrt(pi)

    Another common scaling is via the recurrence
       1
       x
       2 * x * Tn(x) - T{n-1}(x)
    which leads to the equivalence

       Tn(x) = cos(n arccos(x)).

    This isn't implemented here. It can be retrieved from monic scaling by multiplying
    with 2 ** {n-1}. Perhaps this scaling should be added?
    """

    def __init__(self, X, scaling: str, symbolic="auto"):
        if symbolic == "auto":
            symbolic = np.asarray(X).dtype == sympy.Basic

        lmbda = -sympy.S(1) / 2 if symbolic else -0.5
        self._gegenbauer_eval = gegenbauer.Eval(X, scaling, lmbda, symbolic)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._gegenbauer_eval)



from __future__ import annotations

import math

try:
    # Python 3.8+
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import numpy as np
import sympy

from ..helpers import Eval1D


class Eval:
    """Generalized Laguerre polynomials. Set alpha=0 (default) to get classical
    Laguerre.

    The first few are (for alpha=0):

    scaling == "monic":
        1
        x - 1
        x**2 - 4*x + 2
        x**3 - 9*x**2 + 18*x - 6
        x**4 - 16*x**3 + 72*x**2 - 96*x + 24
        x**5 - 25*x**4 + 200*x**3 - 600*x**2 + 600*x - 120

    scaling == "classical" or "normal"
        1
        -x + 1
        x**2/2 - 2*x + 1
        -x**3/6 + 3*x**2/2 - 3*x + 1
        x**4/24 - 2*x**3/3 + 3*x**2 - 4*x + 1
        -x**5/120 + 5*x**4/24 - 5*x**3/3 + 5*x**2 - 5*x + 1

    The classical and normal standarizations differ for alpha != 0.
    """

    def __init__(self, X, *args, symbolic: Literal["auto"] | bool = "auto", **kwargs):
        if symbolic == "auto":
            symbolic = np.asarray(X).dtype == sympy.Basic

        assert isinstance(symbolic, bool)
        rc = RecurrenceCoefficients(*args, symbolic=symbolic, **kwargs)
        self.int_p0 = rc.p0
        self._eval_1d = Eval1D(X, rc)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._eval_1d)


class RecurrenceCoefficients:
    def __init__(
        self,
        scaling: Literal["monic"] | Literal["classical"] | Literal["normal"],
        alpha: int | float = 0,
        symbolic: bool = False,
    ):
        cls = {"monic": RCMonic, "classical": RCClassical, "normal": RCNormal}[scaling]
        self.rc = cls(alpha, symbolic)
        self.p0 = self.rc.p0

        gamma = sympy.gamma if symbolic else lambda x: math.gamma(float(x))
        self.int_1 = gamma(alpha + 1)

    def __getitem__(self, N: int):
        return self.rc[N]


class RCMonic:
    def __init__(self, alpha: int | float, symbolic: bool):
        self.nan = None if symbolic else math.nan
        self.alpha = alpha
        self.p0 = 1

    def __getitem__(self, k: int):
        a = 1
        b = 2 * k + 1 + self.alpha
        c = k * (k + self.alpha) if k > 0 else self.nan
        return a, b, c


class RCClassical:
    def __init__(self, alpha: int | float, symbolic: bool):
        self.nan = None if symbolic else math.nan
        self.S = sympy.S if symbolic else lambda a: a
        self.alpha = alpha
        self.p0 = 1

    def __getitem__(self, k):
        alpha = self.alpha
        S = self.S

        a = -S(1) / (k + 1)
        b = -S(2 * k + 1 + alpha) / (k + 1)
        c = S(k + alpha) / (k + 1) if k > 0 else self.nan
        return a, b, c


class RCNormal:
    def __init__(self, alpha: int | float, symbolic: bool):
        self.nan = None if symbolic else math.nan
        self.sqrt = sympy.sqrt if symbolic else math.sqrt
        self.S = sympy.S if symbolic else lambda a: a
        self.alpha = alpha

        gamma = sympy.gamma if symbolic else math.gamma
        self.p0 = 1 / self.sqrt(gamma(alpha + 1))

    def __getitem__(self, k):
        sqrt = self.sqrt
        S = self.S
        alpha = self.alpha

        a = -1 / sqrt((k + 1) * (k + 1 + alpha))
        b = -(2 * k + 1 + alpha) / sqrt((k + 1) * (k + 1 + alpha))
        c = sqrt(k * S(k + alpha) / ((k + 1) * (k + 1 + alpha))) if k > 0 else self.nan
        return a, b, c



import math

import numpy as np
import sympy


def savefig_single(filename, *args, **kwargs):
    from matplotlib import pyplot as plt

    plot_single(*args, **kwargs)
    plt.savefig(filename, transparent=True, bbox_inches="tight")


def show_single(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot_single(*args, **kwargs)
    plt.show()


def plot_single(*args, **kwargs):
    from .tools import plot_single as ps

    ps("Zernike", Eval, *args, **kwargs)


def savefig_tree(filename, *args, dpi=None, **kwargs):
    from matplotlib import pyplot as plt

    plot_tree(*args, **kwargs)
    plt.savefig(filename, dpi=dpi, transparent=True, bbox_inches="tight")


def show_tree(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot_tree(*args, **kwargs)
    plt.show()


def plot_tree(*args, **kwargs):
    from .tools import plot_tree as pt

    pt("Zernike", Eval, *args, **kwargs)


class Eval:
    """
    Torben B. Andersen,
    Efficient and robust recurrence relations for the Zernike circle polynomials and
    their derivatives in Cartesian coordinates,
    Optics Express Vol. 26, Issue 15, pp. 18878-18896 (2018),
    <https://doi.org/10.1364/OE.26.018878>.
    """

    def __init__(self, X, scaling, symbolic="auto"):
        if symbolic == "auto":
            symbolic = np.asarray(X).dtype == sympy.Basic

        self.rc = {"classical": RCClassical, "normal": RCNormal}[scaling](symbolic)

        self.X = X
        self.L = 0
        self.last = [None, None]

        pi = sympy.pi if symbolic else np.pi
        self.int_p0 = self.rc.p0 * pi

    def __iter__(self):
        return self

    def __next__(self):
        if self.L == 0:
            out = np.array([0 * self.X[0] + self.rc.p0])
        else:
            alpha, beta, gamma = self.rc[self.L]

            shape = list(self.last[0].shape)
            shape[0] += 1
            out = np.zeros(shape, dtype=self.last[0].dtype)

            n = self.L + 1
            half = n // 2

            if n % 2 == 0 and n > 2:
                self.last[0][half - 1] *= beta

            last_X = alpha * self.last[0] * self.X[0]
            last_Y = alpha * self.last[0] * self.X[1]

            out[:-1] += last_X + last_Y[::-1]
            out[1:] += last_X - last_Y[::-1]

            # It works without this seam correction, too. See zernike2.
            if n % 2 == 0:
                out[half - 1] -= last_X[half - 1]
                out[half] += last_Y[half - 1]
            else:
                out[half] += last_X[half]
                out[half] += last_Y[half - 1]

            if self.L > 1:
                out[1:-1] -= gamma * self.last[1]

            if n % 2 == 1:
                out[half] *= 1 / beta

        self.last[1] = self.last[0]
        self.last[0] = out
        self.L += 1
        return out


class RCClassical:
    def __init__(self, symbolic):
        self.p0 = 1

    def __getitem__(self, n):
        assert n > 0

        alpha = 1
        beta = 1
        gamma = None if n == 1 else 1

        return alpha, beta, gamma


class RCNormal:
    def __init__(self, symbolic):
        self.S = sympy.S if symbolic else lambda x: x
        self.sqrt = sympy.sqrt if symbolic else math.sqrt
        pi = sympy.pi if symbolic else math.pi
        self.p0 = 1 / self.sqrt(pi)

    def __getitem__(self, n):
        assert n > 0
        n = self.S(n)

        beta = self.sqrt(2)
        if n == 1:
            alpha = self.sqrt(2 * (n + 1) / n)
            gamma = None
        elif n == 2:
            alpha = self.sqrt((n + 1) / n)
            gamma = self.sqrt(2 * (n + 1) / (n - 1))
        else:
            alpha = self.sqrt((n + 1) / n)
            gamma = self.sqrt((n + 1) / (n - 1))

        return alpha, beta, gamma



from . import associated_legendre, chebyshev1, chebyshev2, gegenbauer, jacobi, legendre
from .tools import clenshaw

__all__ = [
    "associated_legendre",
    "chebyshev1",
    "chebyshev2",
    "gegenbauer",
    "jacobi",
    "legendre",
    "clenshaw",
]



import numpy as np
import sympy

from ..e1r2.main import RCPhysicistNormal, RCProbabilistNormal
from ..helpers import ProductEval, ProductEvalWithDegrees


class Eval:
    def __init__(self, X, standardization, symbolic="auto", return_degrees=False):
        if symbolic == "auto":
            symbolic = np.asarray(X).dtype == sympy.Basic

        rc = {"probabilists": RCProbabilistNormal, "physicists": RCPhysicistNormal}[
            standardization
        ](symbolic)

        sqrt = sympy.sqrt if symbolic else np.sqrt
        pi = sympy.pi if symbolic else np.pi
        int_1 = sqrt(pi) if standardization == "physicists" else 1

        cls = ProductEvalWithDegrees if return_degrees else ProductEval
        self._product_eval = cls(rc, int_1, X)
        self.int_p0 = self._product_eval.int_p0

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._product_eval)



import math

import numpy as np
import sympy

from ..helpers import Eval1D


class Eval:
    """Recurrence coefficients for Hermite polynomials.

    The physicicist's standardization is with respect to the weight function exp(-x^2),
    the probabilists' standardization is with respect to the weight function
    1/sqrt(2*pi) * exp(-x^2/2). See <https://en.wikipedia.org/wiki/Hermite_polynomials>.

    The first few are:
    Probabilist:
        scaling in ["monic", "classical"]:
            1
            x
            x**2 - 1
            x**3 - 3*x
            x**4 - 6*x**2 + 3
            x**5 - 10*x**3 + 15*x

        scaling == "normal":
            1
            x
            sqrt(2)*x**2/2 - sqrt(2)/2
            sqrt(6)*x**3/6 - sqrt(6)*x/2
            sqrt(6)*x**4/12 - sqrt(6)*x**2/2 + sqrt(6)/4
            sqrt(30)*x**5/60 - sqrt(30)*x**3/6 + sqrt(30)*x/4

    Physicist:
        scaling == "monic":
            1
            x
            x**2 - 1/2
            x**3 - 3*x/2
            x**4 - 3*x**2 + 3/4
            x**5 - 5*x**3 + 15*x/4

        scaling == "classical":
            1
            2*x
            4*x**2 - 2
            8*x**3 - 12*x
            16*x**4 - 48*x**2 + 12
            32*x**5 - 160*x**3 + 120*x

        scaling == "normal":
            pi**(-1/4)
            sqrt(2)*x/pi**(1/4)
            sqrt(2)*x**2/pi**(1/4) - sqrt(2)/(2*pi**(1/4))
            2*sqrt(3)*x**3/(3*pi**(1/4)) - sqrt(3)*x/pi**(1/4)
            sqrt(6)*x**4/(3*pi**(1/4)) - sqrt(6)*x**2/pi**(1/4) + sqrt(6)/(4*pi**(1/4))
            2*sqrt(15)*x**5/(15*pi**(1/4)) - 2*sqrt(15)*x**3/(3*pi**(1/4)) + sqrt(15)*x/(2*pi**(1/4))
    """

    def __init__(self, X, *args, symbolic="auto", **kwargs):
        if symbolic == "auto":
            symbolic = np.asarray(X).dtype == sympy.Basic

        rc = RecurrenceCoefficients(*args, symbolic=symbolic, **kwargs)
        self.int_p0 = rc.p0 * rc.int_1
        self._eval_1d = Eval1D(X, rc)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._eval_1d)


class RecurrenceCoefficients:
    def __init__(self, standardization, scaling, symbolic):
        self.rc = {
            "probabilists": {
                # The classical scheme is monic
                "monic": RCProbabilistMonic,
                "classical": RCProbabilistMonic,
                "normal": RCProbabilistNormal,
            },
            "physicists": {
                "monic": RCPhysicistMonic,
                "classical": RCPhysicistClassical,
                "normal": RCPhysicistNormal,
            },
        }[standardization][scaling](symbolic)
        self.p0 = self.rc.p0

        sqrt = sympy.sqrt if symbolic else math.sqrt
        pi = sympy.pi if symbolic else math.pi
        self.int_1 = 1 if standardization == "probabilists" else sqrt(pi)

    def __getitem__(self, N):
        return self.rc[N]


class RCProbabilistMonic:
    def __init__(self, symbolic):
        self.nan = None if symbolic else math.nan
        self.p0 = 1

    def __getitem__(self, k):
        a = 1
        b = 0
        c = k if k > 0 else self.nan
        return a, b, c


class RCProbabilistNormal:
    def __init__(self, symbolic):
        self.frac = sympy.Rational if symbolic else lambda a, b: a / b
        self.nan = None if symbolic else math.nan
        self.sqrt = sympy.sqrt if symbolic else math.sqrt
        self.p0 = 1

    def __getitem__(self, k):
        a = 1 / self.sqrt(k + 1)
        b = 0
        c = self.sqrt(self.frac(k, k + 1)) if k > 0 else self.nan
        return a, b, c


class RCPhysicistMonic:
    def __init__(self, symbolic):
        self.frac = sympy.Rational if symbolic else lambda a, b: a / b
        self.nan = None if symbolic else math.nan
        self.p0 = 1

    def __getitem__(self, k):
        a = 1
        b = 0
        c = self.frac(k, 2) if k > 0 else self.nan
        return a, b, c


class RCPhysicistClassical:
    def __init__(self, symbolic):
        self.nan = None if symbolic else math.nan
        self.p0 = 1

    def __getitem__(self, k):
        a = 2
        b = 0
        c = 2 * k if k > 0 else self.nan
        return a, b, c


class RCPhysicistNormal:
    def __init__(self, symbolic):
        self.frac = sympy.Rational if symbolic else lambda a, b: a / b
        self.nan = None if symbolic else math.nan
        self.sqrt = sympy.sqrt if symbolic else math.sqrt

        pi = sympy.pi if symbolic else math.pi
        self.p0 = 1 / self.sqrt(self.sqrt(pi))

    def __getitem__(self, k):
        a = self.sqrt(self.frac(2, k + 1))
        b = 0
        c = self.sqrt(self.frac(k, k + 1)) if k > 0 else self.nan
        return a, b, c



from .main import Eval
from .tools import (
    plot_tree_1d,
    plot_tree_2d,
    savefig_tree_1d,
    savefig_tree_2d,
    show_tree_1d,
    show_tree_2d,
    write_tree_3d,
)

__all__ = [
    "Eval",
    "plot_tree_1d",
    "show_tree_1d",
    "savefig_tree_1d",
    "plot_tree_2d",
    "show_tree_2d",
    "savefig_tree_2d",
    "write_tree_3d",
]



import itertools

import numpy as np

from .main import Eval


def plot(n, *args, **kwargs):
    import matplotx
    from matplotlib import pyplot as plt

    plt.style.use(matplotx.styles.dufte)

    x = np.linspace(-2.2, 2.2, 100)
    for k, level in enumerate(itertools.islice(Eval(x, *args, **kwargs), n + 1)):
        plt.plot(x, level, label=f"n={k}")

    plt.grid(axis="x")
    matplotx.line_labels()

    variant, scaling = args
    plt.title(f"Hermite polynomials ({variant}, scaling={scaling})")


def show(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot(*args, **kwargs)
    plt.show()


def savefig(filename, *args, **kwargs):
    from matplotlib import pyplot as plt

    plot(*args, **kwargs)
    plt.savefig(filename, transparent=True, bbox_inches="tight")



from . import jacobi


def plot(n, scaling, lmbda):
    jacobi.plot(n, scaling, lmbda, lmbda)


def show(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot(*args, **kwargs)
    plt.show()


def savefig(filename, *args, **kwargs):
    from matplotlib import pyplot as plt

    plot(*args, **kwargs)
    plt.savefig(filename, transparent=True, bbox_inches="tight")


class Eval:
    def __init__(self, X, scaling, lmbda, symbolic="auto"):
        self._jacobi_eval = jacobi.Eval(X, scaling, lmbda, lmbda, symbolic=symbolic)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._jacobi_eval)


class RecurrenceCoefficients:
    def __init__(self, scaling, lmbda, symbolic="auto"):
        self._jacobi_rc = jacobi.RecurrenceCoefficients(
            scaling, lmbda, lmbda, symbolic=symbolic
        )
        self.int_1 = self._jacobi_rc.int_1

    def __getitem__(self, N):
        return self._jacobi_rc[N]



import numpy as np
import sympy


def full_like(x, val):
    if isinstance(x, np.ndarray):
        return np.full_like(x, val)

    # assume x is just a float or int or sympy.Poly
    return x * 0 + val


class Eval1D:
    def __init__(self, x, rc):
        self.rc = rc
        self.x = x
        self.k = 0
        self.last = [None, None]

    def __iter__(self):
        return self

    def __next__(self):
        if self.k == 0:
            out = full_like(self.x, self.rc.p0)
        else:
            a, b, c = self.rc[self.k - 1]
            out = self.last[0] * (self.x * a - b)
            if self.k > 1:
                out -= self.last[1] * c

        self.last[1] = self.last[0]
        self.last[0] = out
        self.k += 1
        return out


class ProductEvalWithDegrees:
    """Evaluates the entire tree of orthogonal polynomials for an n-dimensional product
    domain.

    The computation is organized such that tree returns a list of arrays, L={0, ...,
    dim}, where each level corresponds to the polynomial degree L. Further, each level
    is organized like a discrete (dim-1)-dimensional simplex. Let's demonstrate this for
    3D:

    L = 1:
                   (0, 0, 0)

    L = 2:
                   (1, 0, 0)
              (0, 1, 0) (0, 0, 1)

    L = 3:
                   (2, 0, 0)
              (1, 1, 0) (1, 0, 1)
         (0, 2, 0) (0, 1, 1) (0, 0, 2)

    L = 4:
                   (3, 0, 0)
              (2, 1, 0) (2, 0, 1)
         (1, 2, 0) (1, 1, 1) (1, 0, 2)
    (0, 3, 0) (0, 2, 1) (0, 1, 2) (0, 0, 3)

    The main insight here that makes computation for n dimensions easy is that the next
    level is composed by:

       * Taking the whole previous level and adding +1 to the first entry.
       * Taking the last row of the previous level and adding +1 to the second entry.
       * Taking the last entry of the last row of the previous and adding +1 to the
         third entry.

    In the same manner this can be repeated for `dim` dimensions.
    """

    def __init__(self, rc, int_1, X):
        self.rc = rc

        self.a = None
        self.b = None
        self.c = None
        X = np.asarray(X)
        self.dim = X.shape[0]
        self.p0n = rc.p0 ** self.dim
        self.int_p0 = self.p0n * int_1 ** self.dim
        self.L = 0
        self.X = X
        self.last_values = [None, None]
        self.last_degrees = [None, None]

    def __iter__(self):
        return self

    def __next__(self):
        X = self.X
        dim = X.shape[0]

        if self.L == 0:
            values = np.array([X[0] * 0 + self.p0n])
            degrees = np.array([np.zeros(dim, dtype=int)])
        else:
            aa, bb, cc = self.rc[self.L - 1]
            # cannot just np.append here since numpy will convert to float64
            # https://github.com/numpy/numpy/issues/18189
            self.a = np.array([aa]) if self.a is None else np.append(self.a, aa)
            self.b = np.array([bb]) if self.b is None else np.append(self.b, bb)
            self.c = np.array([cc]) if self.c is None else np.append(self.c, cc)

            a = self.a
            b = self.b
            c = self.c

            values = []
            degrees = []

            mask0 = np.ones(len(self.last_degrees[0]), dtype=bool)
            if self.L > 1:
                mask1 = np.ones(len(self.last_degrees[1]), dtype=bool)

            for i in range(dim):
                lv0 = self.last_values[0][mask0]
                idx0 = self.last_degrees[0][mask0][:, i]

                val = lv0 * (np.multiply.outer(a[idx0], X[i]).T - b[idx0]).T

                if self.L > 1:
                    lv1 = self.last_values[1][mask1]
                    idx1 = self.last_degrees[1][mask1][:, i]
                    yy = idx1 + 1 > 0
                    val[: len(idx1)][yy] -= (lv1[yy].T * c[idx1[yy] + 1]).T

                values.append(val)

                deg = self.last_degrees[0][mask0]
                deg[:, i] += 1
                degrees.append(deg)
                # mask is True for all entries where the first `i` degrees are 0
                mask0 &= self.last_degrees[0][:, i] == 0
                if self.L > 1:
                    mask1 &= self.last_degrees[1][:, i] == 0

            values = np.concatenate(values)
            degrees = np.concatenate(degrees)

        self.last_values[1] = self.last_values[0]
        self.last_values[0] = values

        self.last_degrees[1] = self.last_degrees[0]
        self.last_degrees[0] = degrees
        self.L += 1

        return values, degrees


class ProductEval(ProductEvalWithDegrees):
    """Same as ProductEvalWithDegrees, but next() only returns the values."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __next__(self):
        vals, _ = super().__next__()
        return vals


class Eval135:
    """Evaluates a 1-3-5-tree as seen with associated Legendre polynomials and spherical
    harmonics.

    There are many recurrence relations that can be used to construct the associated
    Legendre polynomials. However, only few are numerically stable. Many implementations
    (including this one) use the classical Legendre recurrence relation with increasing
    L.

    The return value is a list of arrays, where `values[k]` hosts the `2*k+1` values of
    the `k`th level of the tree

                              (0, 0)
                    (-1, 1)   (0, 1)   (1, 1)
          (-2, 2)   (-1, 2)   (0, 2)   (1, 2)   (2, 2)
            ...       ...       ...     ...       ...
    """

    def __init__(self, rc, x, xi=None, symbolic=False):
        self.rc = rc

        self.k = 0
        self.x = x
        # xi[0] == sqrt(1 - x**2) / exp(i*phi)
        # xi[1] == sqrt(1 - x**2) * exp(i*phi)
        if xi is None:
            sqrt = np.vectorize(sympy.sqrt) if symbolic else np.sqrt
            # Such functions aren't always polynomials, see, e.g.,
            # <https://en.wikipedia.org/wiki/Associated_Legendre_polynomials>:
            #
            # > In general, when l and m are integers, the regular solutions are
            # > sometimes called "associated Legendre polynomials", even though they are
            # > not polynomials when m is odd.
            a = sqrt(1 - x ** 2)
            self.xi = [a, a]
        else:
            self.xi = xi

        self.last = [None, None]

    def __iter__(self):
        return self

    def __next__(self):
        if self.k == 0:
            out = np.array([full_like(self.x, self.rc.p0)])
        else:
            z0, z1, c0, c1 = self.rc[self.k]
            out = np.concatenate(
                [
                    [self.last[0][0] * (self.xi[0] * z0)],
                    self.last[0] * np.multiply.outer(c0, self.x),
                    [self.last[0][-1] * (self.xi[1] * z1)],
                ]
            )

            if self.k > 1:
                out[2:-2] -= (self.last[1].T * c1).T

        self.last[1] = self.last[0]
        self.last[0] = out
        self.k += 1
        return out



import numpy as np
import sympy

from ..helpers import Eval135


class EvalCartesian:
    """Evaluate spherical harmonics degree by degree `n` at angles `polar`, `azimuthal`."""

    def __init__(self, X, scaling, complex_valued=True, symbolic="auto"):
        assert len(X) == 3
        # assert X[0] ** 2 + X[1] ** 2 + X[2] ** 2 == 1

        if symbolic == "auto":
            symbolic = np.asarray(X).dtype == sympy.Basic

        # Conventions from
        # <https://en.wikipedia.org/wiki/Spherical_harmonics#Orthogonality_and_normalization>.
        rc = {
            "acoustic": RCSpherical(False, symbolic, geodetic=False),
            "quantum mechanic": RCSpherical(True, symbolic, geodetic=False),
            "geodetic": RCSpherical(False, symbolic, geodetic=True),
            "schmidt": RCSchmidt(False, symbolic),
        }[scaling]

        if complex_valued:
            #
            # x = sin(polar) * cos(azimuthal)
            # y = sin(polar) * sin(azimuthal)
            # z = cos(polar)
            #
            # exp_iphi
            # = exp(imag_unit * azimuthal)
            # = (X[0] + imag_unit * X[1]) / sqrt(X[0] ** 2 + X[1] ** 2)
            #
            # sqrt(1 - X[2]**2) / exp(i * phi)
            # = (X[0] ** 2 + X[1] ** 2) / (X[0] + imag_unit * X[1])
            # = X[0] - imag_unit * X[1]
            #
            # sqrt(1 - X[2]**2) * exp(i * phi)
            # = X[0] + imag_unit * X[1]
            #
            imag_unit = sympy.I if symbolic else 1j
            xi = [
                X[0] - imag_unit * X[1],
                X[0] + imag_unit * X[1],
            ]
        else:
            # real-valued, but not always a polynomial; see associated Legendre
            # functions
            sqrt = sympy.sqrt if symbolic else np.sqrt
            a = sqrt(X[0] ** 2 + X[1] ** 2)
            xi = [a, a]

        pi = sympy.pi if symbolic else np.pi
        self.int_p0 = rc.p0 * 4 * pi

        self._eval_135 = Eval135(rc, X[2], xi, symbolic=symbolic)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._eval_135)


class EvalSpherical:
    """Evaluate spherical harmonics degree by degree `n` at angles `polar`, `azimuthal`."""

    def __init__(self, theta_phi, scaling, complex_valued=True, symbolic="auto"):
        assert len(theta_phi) == 2
        if symbolic == "auto":
            symbolic = np.asarray(theta_phi).dtype == sympy.Basic

        # Conventions from
        # <https://en.wikipedia.org/wiki/Spherical_harmonics#Orthogonality_and_normalization>.
        rc = {
            "acoustic": RCSpherical(False, symbolic, geodetic=False),
            "quantum mechanic": RCSpherical(True, symbolic, geodetic=False),
            "geodetic": RCSpherical(False, symbolic, geodetic=True),
            "schmidt": RCSchmidt(False, symbolic),
        }[scaling]

        sin = np.vectorize(sympy.sin) if symbolic else np.sin
        cos = np.vectorize(sympy.cos) if symbolic else np.cos

        # X = [
        #     sin_polar * cos_azimu,
        #     sin_polar * sin_azimu,
        #     cos_polar,
        # ]

        if complex_valued:
            sin_theta, sin_phi = sin(theta_phi)
            cos_theta, cos_phi = cos(theta_phi)

            imag_unit = sympy.I if symbolic else 1j
            exp_i_phi = cos_phi + imag_unit * sin_phi
            xi = [sin_theta / exp_i_phi, sin_theta * exp_i_phi]
        else:
            theta = theta_phi[0]
            sin_theta = sin(theta)
            cos_theta = cos(theta)
            xi = [sin_theta, sin_theta]

        self._eval_135 = Eval135(rc, cos_theta, xi, symbolic=symbolic)

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._eval_135)


class RCSpherical:
    def __init__(self, with_cs_phase, symbolic, geodetic):
        pi = sympy.pi if symbolic else np.pi
        self.sqrt = np.vectorize(sympy.sqrt) if symbolic else np.sqrt
        self.S = sympy.S if symbolic else lambda x: x

        # The starting value 1 has the effect of multiplying the entire tree by
        # sqrt(4*pi). This convention is used in geodesy and and spectral
        # analysis.
        self.p0 = 1 if geodetic else 1 / self.sqrt(4 * pi)
        self.phase = -1 if with_cs_phase else 1

    def __getitem__(self, L):
        L = self.S(L)

        z0 = self.sqrt((2 * L + 1) / (2 * L))
        z1 = z0 * self.phase
        #
        m = np.arange(-L + 1, L)
        c0 = self.sqrt((2 * L - 1) * (2 * L + 1) / ((L + m) * (L - m)))
        #
        if L == 1:
            c1 = None
        else:
            m = np.arange(-L + 2, L - 1)
            c1 = self.sqrt(
                (L + m - 1)
                * (L - m - 1)
                * (2 * L + 1)
                / ((2 * L - 3) * (L + m) * (L - m))
            )
        return z0, z1, c0, c1


class RCSchmidt:
    def __init__(self, with_cs_phase, symbolic):
        self.sqrt = np.vectorize(sympy.sqrt) if symbolic else np.sqrt
        self.S = sympy.S if symbolic else lambda x: x
        self.phase = -1 if with_cs_phase else 1

        # if phi is None:
        #     self.p0 = 2
        self.p0 = 1

    def __getitem__(self, L):
        L = self.S(L)

        z0 = self.sqrt((2 * L - 1) / (2 * L))
        z1 = z0 * self.phase
        #
        m = np.arange(-L + 1, L)
        c0 = (2 * L - 1) / self.sqrt((L + m) * (L - m))
        #
        if L == 1:
            c1 = None
        else:
            m = np.arange(-L + 2, L - 1)
            c1 = self.sqrt((L + m - 1) * (L - m - 1) / ((L + m) * (L - m)))

        return z0, z1, c0, c1



"""
[1] Gene H. Golub and John H. Welsch,
    Calculation of Gauss Quadrature Rules,
    Mathematics of Computation,
    Vol. 23, No. 106 (Apr., 1969), pp. 221-230+s1-s10,
    <https://dx.doi.org/10.2307/2004418>,
    <https://pdfs.semanticscholar.org/c715/119d5464f614fd8ec590b732ccfea53e72c4.pdf>.

[2] W. Gautschi,
    Algorithm 726: ORTHPOL–a package of routines for generating orthogonal polynomials
    and Gauss-type quadrature rules,
    ACM Transactions on Mathematical Software (TOMS),
    Volume 20, Issue 1, March 1994,
    Pages 21-62,
    <https://doi.org/10.1145/174603.174605>,
    <https://www.cs.purdue.edu/archives/2002/wxg/codes/gauss.m>,

[3] W. Gautschi,
    How and how not to check Gaussian quadrature formulae,
    BIT Numerical Mathematics,
    June 1983, Volume 23, Issue 2, pp 209–216,
    <https://doi.org/10.1007/BF02218441>.
"""
import math

import numpy as np
import sympy


def golub_welsch(moments):
    """Given moments

    mu_k = int_a^b omega(x) x^k dx,  k = {0, 1,...,2N}

    (with omega being a nonnegative weight function), this method creates the recurrence
    coefficients of the corresponding orthogonal polynomials, see section 4
    ("Determining the Three Term Relationship from the Moments") in Golub-Welsch [1].
    Numerically unstable, see [2].
    """
    assert len(moments) % 2 == 1
    n = (len(moments) - 1) // 2

    M = np.array([[moments[i + j] for j in range(n + 1)] for i in range(n + 1)])
    R = np.linalg.cholesky(M).T

    # (upper) diagonal
    Rd = R.diagonal()
    q = R.diagonal(1) / Rd[:-1]

    alpha = q.copy()
    alpha[+1:] -= q[:-1]

    beta = np.hstack([math.nan, Rd[1:-1] / Rd[:-2]]) ** 2
    int_1 = Rd[0]
    return alpha, beta, int_1


# We could make stieltjes() an iterator with __next__() easily enough, but the same
# isn't possible to chebyshev(). For the sake of consistency, keep it a function.
def stieltjes(integrate, n):
    t = sympy.Symbol("t")

    alpha = n * [None]
    beta = n * [None]
    mu = [None, None]
    # See <https://github.com/microsoft/pyright/issues/1229> for the pyright issues.
    pi = [None, None, None]
    int_1 = None

    for k in range(n):
        pi[1], pi[2] = pi[0], pi[1]
        mu[1] = mu[0]

        if k == 0:
            pi[0] = 1
            mu[0] = integrate(t, pi[0] ** 2)
            alpha[k] = integrate(t, t * pi[0] ** 2) / mu[0]
            int_1 = mu[0]
        else:
            pi[0] = (t - alpha[k - 1]) * pi[1]
            if k > 1:
                pi[0] -= beta[k - 1] * pi[2]

            mu[0] = integrate(t, pi[0] ** 2)
            alpha[k] = integrate(t, t * pi[0] ** 2) / mu[0]
            beta[k] = mu[0] / mu[1]

    return alpha, beta, int_1


def chebyshev(moments):
    """Given the first 2n moments `int t^k dt`, this method uses the Chebyshev algorithm
    (see, e.g., [2]) for computing the associated recurrence coefficients.

    WARNING: Ill-conditioned, see [2].
    """
    m = len(moments)
    assert m % 2 == 0
    # https://stackoverflow.com/a/30039361/353337
    dtype = sympy.Rational if isinstance(moments[0], sympy.Basic) else moments.dtype
    zeros = np.zeros((m, 3), dtype=dtype)
    return chebyshev_modified(moments, zeros)


def chebyshev_modified(nu, recurrence_coefficients):
    """Given the first 2n modified moments `nu_k = int p_k(t) dt`, where the p_k are
    orthogonal polynomials with recurrence coefficients a, b, this method implements the
    modified Chebyshev algorithm (see, e.g., [2]) for computing the associated
    recurrence coefficients.
    """
    m = len(nu)
    assert m % 2 == 0, "Need an even number of moments."

    n = m // 2

    alpha = []
    beta = []
    sigma = [None, None, None]
    int_1 = nu[0]

    if n > 0:
        k = 0
        sigma[0] = np.asarray(nu)
        _, a0, _ = recurrence_coefficients[0]
        alpha.append(a0 + nu[1] / nu[0])
        beta.append(math.nan)

    for k in range(1, n):
        sigma[2], sigma[1] = sigma[1], sigma[0]

        _, aL, bL = np.array(
            [recurrence_coefficients[i] for i in range(k, 2 * n - k)]
        ).T
        sigma[0] = (
            sigma[0][2:] - (alpha[k - 1] - aL) * sigma[0][1:-1] + bL * sigma[1][:-2]
        )
        if k > 1:
            sigma[0] -= beta[k - 1] * sigma[2][2:-2]

        ak = aL[0]
        alpha.append(ak + sigma[0][1] / sigma[0][0] - sigma[1][1] / sigma[1][0])
        beta.append(sigma[0][0] / sigma[1][0])

    return np.asarray(alpha), np.asarray(beta), int_1


def gautschi_test_3(moments, alpha, beta):
    """In his article [3], Walter Gautschi suggests a method for checking if a
    quadrature rule is sane. This method implements test #3 for the article.
    """
    n = len(alpha)
    assert len(beta) == n

    D = np.empty(n + 1)
    Dp = np.empty(n + 1)
    D[0] = 1.0
    D[1] = moments[0]
    Dp[0] = 0.0
    Dp[1] = moments[1]
    for k in range(2, n + 1):
        A = np.array([moments[i : i + k] for i in range(k)])
        D[k] = np.linalg.det(A)
        #
        A[:, -1] = moments[k : 2 * k]
        Dp[k] = np.linalg.det(A)

    errors_alpha = np.zeros(n)
    errors_beta = np.zeros(n)

    errors_alpha[0] = abs(alpha[0] - (Dp[1] / D[1] - Dp[0] / D[0]))
    errors_beta[0] = abs(beta[0] - D[1])
    for k in range(1, n):
        errors_alpha[k] = abs(alpha[k] - (Dp[k + 1] / D[k + 1] - Dp[k] / D[k]))
        errors_beta[k] = abs(beta[k] - D[k + 1] * D[k - 1] / D[k] ** 2)

    return errors_alpha, errors_beta



from __future__ import annotations

try:
    # Python 3.8+
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import math

import numpy as np
import sympy
from numpy.typing import ArrayLike

from .tools import plot_single as ps


def savefig_single(filename, *args, **kwargs):
    from matplotlib import pyplot as plt

    plot_single(*args, **kwargs)
    plt.savefig(filename, transparent=True, bbox_inches="tight")


def show_single(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot_single(*args, **kwargs)
    plt.show()


def plot_single(*args, **kwargs):
    ps("Zernike-2", Eval, *args, **kwargs)


def savefig_tree(filename, *args, **kwargs):
    from matplotlib import pyplot as plt

    plot_tree(*args, **kwargs)
    plt.savefig(filename, transparent=True, bbox_inches="tight")


def show_tree(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot_tree(*args, **kwargs)
    plt.show()


def plot_tree(*args, **kwargs):
    from .tools import plot_tree as pt

    pt("Zernike-2", Eval, *args, **kwargs)


class Eval:
    """
    Similar to regular Zernike, but a lot simpler. Can probably be generalized to
    n-ball.
    """

    def __init__(
        self,
        X: ArrayLike,
        scaling: Literal["classical"] | Literal["monic"] | Literal["normal"],
        symbolic: Literal["auto"] | bool = "auto",
    ):
        if symbolic == "auto":
            symbolic = np.asarray(X).dtype == sympy.Basic

        self.rc = {"classical": RCClassical, "monic": RCMonic, "normal": RCNormal}[
            scaling
        ](symbolic)
        self.X = X
        self.L = 0
        self.last = [None, None]

        pi = sympy.pi if symbolic else np.pi
        self.int_p0 = self.rc.p0 * pi

    def __iter__(self):
        return self

    def __next__(self):
        if self.L == 0:
            out = np.array([0 * self.X[0] + self.rc.p0])
        else:
            alpha, beta, gamma = self.rc[self.L]

            shape = list(self.last[0].shape)
            shape[0] += 1
            out = np.zeros(shape, dtype=self.last[0].dtype)

            last_X = self.last[0] * self.X[0]
            last_Y = self.last[0] * self.X[1]

            # The minus sign could go onto the other last_Y, too.
            out[1:] += alpha * (last_X + last_Y[::-1])
            out[:-1] += beta * (last_X - last_Y[::-1])
            if self.L > 1:
                out[1:-1] -= gamma * self.last[1]

        self.last[1] = self.last[0]
        self.last[0] = out
        self.L += 1
        return out


class RCClassical:
    def __init__(self, _):
        self.p0 = 1

    def __getitem__(self, n: int):
        assert n > 0

        alpha = 1
        beta = 1
        if n == 1:
            gamma = None
        else:
            gamma = 1
        return alpha, beta, gamma


class RCMonic:
    def __init__(self, symbolic: bool):
        self.S = sympy.S if symbolic else lambda x: x
        self.p0 = 1

    def __getitem__(self, n: int):
        assert n > 0

        n = self.S(n)
        alpha = np.array([i / n for i in range(1, n + 1)])
        beta = np.array([(n - i) / n for i in range(n)])
        if n == 1:
            gamma = None
        else:
            gamma = np.array([i * (n - i) / (n * (n - 1)) for i in range(1, n)])
        return alpha, beta, gamma


class RCNormal:
    def __init__(self, symbolic: bool):
        self.S = sympy.S if symbolic else lambda x: x
        self.sqrt = sympy.sqrt if symbolic else math.sqrt
        pi = sympy.pi if symbolic else math.pi
        self.p0 = 1 / self.sqrt(pi)

    def __getitem__(self, n: int):
        assert n > 0
        n = self.S(n)
        sqrt = self.sqrt

        alpha = sqrt((n + 1) / n)
        beta = sqrt((n + 1) / n)
        if n == 1:
            gamma = None
        else:
            gamma = sqrt((n + 1) / (n - 1))
        return alpha, beta, gamma



import itertools

import numpy as np

from ..e1r2 import plot as plot_hermite
from .main import Eval


def show_tree_1d(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot_tree_1d(*args, **kwargs)
    plt.show()
    plt.close()


def savefig_tree_1d(filename, *args, **kwargs):
    from matplotlib import pyplot as plt

    plot_tree_1d(*args, **kwargs)
    plt.savefig(filename, transparent=True, bbox_inches="tight")
    plt.close()


def plot_tree_1d(n, *args, **kwargs):
    plot_hermite(n, *args, **kwargs)


def show_tree_2d(*args, **kwargs):
    from matplotlib import pyplot as plt

    plot_tree_2d(*args, **kwargs)
    plt.show()
    plt.close()


def savefig_tree_2d(filename, *args, **kwargs):
    from matplotlib import pyplot as plt

    plot_tree_2d(*args, **kwargs)
    plt.savefig(filename, transparent=True, bbox_inches="tight")
    plt.close()


def plot_tree_2d(
    n, *args, res=100, alpha=2.0, colorbar=True, cmap="RdBu_r", clim=None, **kwargs
):
    import matplotx
    import meshzoo
    from matplotlib import pyplot as plt

    plt.style.use(matplotx.styles.dufte)

    points, cells = meshzoo.rectangle(-alpha, alpha, -alpha, alpha, res, res)
    evaluator = Eval(points.T, *args, return_degrees=True, **kwargs)

    plt.set_cmap(cmap)
    plt.gca().set_aspect("equal")
    plt.axis("off")

    for k, (values, _) in enumerate(itertools.islice(evaluator, n + 1)):
        for r, z in enumerate(values):
            offset = [alpha * 2.8 * (r - k / 2), -alpha * 2.6 * k]
            pts = points + offset

            plt.tripcolor(pts[:, 0], pts[:, 1], cells, z, shading="flat")
            plt.clim(clim)

            # rectangle outlines
            corners = np.array(
                [
                    [-alpha, alpha, alpha, -alpha, -alpha],
                    [-alpha, -alpha, alpha, alpha, -alpha],
                ]
            )
            corners = (corners.T + offset).T
            plt.plot(corners[0], corners[1], "-k")

    if colorbar:
        plt.colorbar()


def write_tree_3d(filename, n, *args, res=20, alpha=2.0, **kwargs):
    import meshio
    import meshzoo

    points, cells = meshzoo.cube(
        -alpha, alpha, -alpha, alpha, -alpha, alpha, res, res, res
    )

    evaluator = Eval(points.T, *args, return_degrees=True, **kwargs)
    meshes = []

    corners = np.array(
        [[np.cos(k * 2 * np.pi / 3), np.sin(k * 2 * np.pi / 3)] for k in range(3)]
    )

    for L, (values, degrees) in enumerate(itertools.islice(evaluator, n)):
        for vals, degs in zip(values, degrees):
            offset = sum([corner * d for corner, d in zip(corners, degs)])
            pts = points.copy()
            pts[:, 0] += alpha * 2.0 * offset[0]
            pts[:, 1] += alpha * 2.0 * offset[1]
            pts[:, 2] -= alpha * 3.0 * L
            meshes.append(meshio.Mesh(pts, {"tetra": cells}, point_data={"f": vals}))

    # merge meshes
    points = np.concatenate([mesh.points for mesh in meshes])
    f_vals = np.concatenate([mesh.point_data["f"] for mesh in meshes])
    #
    cells = []
    k = 0
    for mesh in meshes:
        cells.append(mesh.cells[0].data + k)
        k += mesh.points.shape[0]
    cells = np.concatenate(cells)

    meshio.write_points_cells(
        filename, points, {"tetra": cells}, point_data={"f": f_vals}
    )



from .generate_rc import (
    chebyshev,
    chebyshev_modified,
    gautschi_test_3,
    golub_welsch,
    stieltjes,
)

__all__ = [
    "golub_welsch",
    "stieltjes",
    "chebyshev",
    "chebyshev_modified",
    "gautschi_test_3",
]
